#define _CRT_SECURE_NO_WARNINGS
#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

// screen resolution (higher resolutions dont scale well as more raycasts must be performed)
// note that the voxel size 'h' can also be used to improve fidelity 
const int screen_width = 1024;
const int screen_height = 1024;

// controls movement speed and turning, if you change voxel size a lot, movement speed may need adjustment
float	stepx = 0.05f,
		stepy = 0.05f,
		stepz = 0.05f,
		degleft = 10.0f,
		degup = 10.0f,
		degroll = 10.0f;

// voxel grid cell size, small voxels look better, but make the map smaller
const float h = 0.0006f;


const unsigned long kFar = 2;
const unsigned long kRange = 10;

// scales view matrix (can be used to scale distance fog)
const float d1 = 0.1f;
const float d2 = 0.1f;

const float ddx = 1.0f / screen_width;
const float ddy = 1.0f / screen_height;


#define WM_TICK 1

#ifndef M_PI
#define M_PI 3.1415926535f
#endif



int cxClient;
int cyClient;

unsigned char *vram = NULL;
bool LoadVxl(char* filename);
void DrawVxlRaycast(unsigned char *buffer, int width, int height);
void VoxelLaserKill();

typedef struct
{
	float x, y, z, w;
} Vector3D;

typedef struct
{
	float m[16];
} Matrix;

typedef struct
{
	Vector3D eye;
	Vector3D center;
	Vector3D up;
	Vector3D side;
} Camera;

Matrix Identity = { 
	1.0, 0.0, 0.0, 0.0,
	0.0, 1.0, 0.0, 0.0,
	0.0, 0.0, 1.0, 0.0,
	0.0, 0.0, 0.0, 1.0 };

Camera default_cam = {
	{ 0.0, 0.0, 2.0, 1.0 },
	{ 0.0, 0.0, 1.0, 1.0 },
	{ 0.0, 1.0, 2.0, 1.0 },
	{ 1.0, 0.0, 2.0, 1.0 } };

Camera *player = &default_cam;


struct Voxel;
typedef void(*ProcessVoxelCallback)(long sx, long sy, float fade, Voxel* voxel, long index);
typedef bool(*VisitVoxelCallback)(long sx, long sy, long x, long y, long z, float fade, ProcessVoxelCallback ProcessVoxel);

Vector3D viewport[3];
Vector3D viewport_dxvector;
Vector3D viewport_dyvector;
Vector3D up;
Vector3D front;
Vector3D right;

struct Point3D
{
	double x, y, z;
	Point3D() : x(0.0), y(0.0), z(0.0) {}
};
Point3D ipos, istr, ihei, ifor;

const long kBitArraySize = 33554432; //1024 * 1024 * 256 / 8
unsigned char bitarray[kBitArraySize];
struct Voxel
{
	unsigned long x, y, z, argb;
	Voxel() : x(0), y(0), z(0), argb(0) {}
};
Voxel** colorarray;



const Vector3D boardorigin = { -512 * h, -512 * h, -128 * h, 1.0 };

typedef void(*MovementCallback)(float amount,
	Vector3D* eye,
	Vector3D* center,
	Vector3D* up,
	Vector3D* side);


Vector3D TransformVector(const Vector3D &v, const Matrix &m)
{
	Vector3D result;
	result.x = (m.m[0] * v.x) + (m.m[4] * v.y) + (m.m[8] * v.z) + (m.m[12] * v.w);
	result.y = (m.m[1] * v.x) + (m.m[5] * v.y) + (m.m[9] * v.z) + (m.m[13] * v.w);
	result.z = (m.m[2] * v.x) + (m.m[6] * v.y) + (m.m[10] * v.z) + (m.m[14] * v.w);
	result.w = (m.m[3] * v.x) + (m.m[7] * v.y) + (m.m[11] * v.z) + (m.m[15] * v.w);
	return result;
}

inline float Length(const Vector3D &v)
{
	return sqrtf(v.x*v.x + v.y*v.y + v.z*v.z);
}


Vector3D Normalize(const Vector3D &v)
{
	Vector3D result = v;
	float length = Length(v);
	if (length)
	{
		result.x /= length;
		result.y /= length;
		result.z /= length;
	}
	return result;
}

//Rodrigues' rotation formula
Matrix Rotate(float degrees, const Vector3D& axis)
{
	Matrix I, Q, S, R;
	unsigned int i;
	float radians = (degrees * M_PI) / 180.0f;

	R = S = Q = I = Identity;

	Q.m[0] = axis.x * axis.x;
	Q.m[4] = axis.x * axis.y;
	Q.m[8] = axis.x * axis.z;
	Q.m[1] = axis.x * axis.y;
	Q.m[5] = axis.y * axis.y;
	Q.m[9] = axis.y * axis.z;
	Q.m[2] = axis.x * axis.z;
	Q.m[6] = axis.y * axis.z;
	Q.m[10] = axis.z  *axis.z;

	S.m[0] = 0.0;
	S.m[4] = -axis.z;
	S.m[8] = axis.y;
	S.m[1] = axis.z;
	S.m[5] = 0.0;
	S.m[9] = -axis.x;
	S.m[2] = -axis.y;
	S.m[6] = axis.x;
	S.m[10] = 0.0;

	for (i = 0; i<12; ++i)
	{
		if (i == 3 || i == 7 || i == 11)
			continue;

		I.m[i] *= cosf(radians);
		Q.m[i] *= (1 - cosf(radians));
		S.m[i] *= sinf(radians);
		R.m[i] = I.m[i] + Q.m[i] + S.m[i];
	}

	return R;
}

Matrix Translate(float x, float y, float z)
{
	Matrix T = Identity;

	T.m[12] = x;
	T.m[13] = y;
	T.m[14] = z;
	T.m[15] = 1.0;

	return T;
}

// ============================================================
// FPS-style 6DOF camera originated from Crystal Ball interface
// ============================================================
void MoveZ(float step, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	Vector3D frontaxis;
	frontaxis.x = center->x - eye->x;
	frontaxis.y = center->y - eye->y;
	frontaxis.z = center->z - eye->z;
	frontaxis = Normalize(frontaxis);

	eye->x += step * frontaxis.x;
	eye->y += step * frontaxis.y;
	eye->z += step * frontaxis.z;
	center->x += step * frontaxis.x;
	center->y += step * frontaxis.y;
	center->z += step * frontaxis.z;
	up->x += step * frontaxis.x;
	up->y += step * frontaxis.y;
	up->z += step * frontaxis.z;
	side->x += step * frontaxis.x;
	side->y += step * frontaxis.y;
	side->z += step * frontaxis.z;
}

void MoveY(float step, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	Vector3D upaxis;
	upaxis.x = up->x - eye->x;
	upaxis.y = up->y - eye->y;
	upaxis.z = up->z - eye->z;
	upaxis = Normalize(upaxis);

	eye->x += step * upaxis.x;
	eye->y += step * upaxis.y;
	eye->z += step * upaxis.z;
	center->x += step * upaxis.x;
	center->y += step * upaxis.y;
	center->z += step * upaxis.z;
	up->x += step * upaxis.x;
	up->y += step * upaxis.y;
	up->z += step * upaxis.z;
	side->x += step * upaxis.x;
	side->y += step * upaxis.y;
	side->z += step * upaxis.z;
}

void MoveX(float step, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	Vector3D sideaxis;
	sideaxis.x = side->x - eye->x;
	sideaxis.y = side->y - eye->y;
	sideaxis.z = side->z - eye->z;
	sideaxis = Normalize(sideaxis);

	eye->x += step * sideaxis.x;
	eye->y += step * sideaxis.y;
	eye->z += step * sideaxis.z;
	center->x += step * sideaxis.x;
	center->y += step * sideaxis.y;
	center->z += step * sideaxis.z;
	up->x += step * sideaxis.x;
	up->y += step * sideaxis.y;
	up->z += step * sideaxis.z;
	side->x += step * sideaxis.x;
	side->y += step * sideaxis.y;
	side->z += step * sideaxis.z;
}

void TurnLeft(float degrees, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	//In Crystal Ball interface:
	//eye = eye * rotate(degrees, normalize(up));
	//*eye = TransformVector(*eye, Rotate(degrees, Normalize(up)));

	//Set up axis to rotate around. It's origin will be the origin of world coordinates.
	Vector3D upaxis;
	upaxis.x = up->x - eye->x; //0.0
	upaxis.y = up->y - eye->y; //1.0
	upaxis.z = up->z - eye->z; //0.0
	upaxis = Normalize(upaxis);

	Vector3D oldeye;
	oldeye.x = eye->x;
	oldeye.y = eye->y;
	oldeye.z = eye->z;

	//Translate to the origin of world coordinates
	center->x -= eye->x;
	center->y -= eye->y;
	center->z -= eye->z;
	up->x -= eye->x;
	up->y -= eye->y;
	up->z -= eye->z;
	side->x -= eye->x;
	side->y -= eye->y;
	side->z -= eye->z;
	eye->x -= eye->x; //
	eye->y -= eye->y; // :)
	eye->z -= eye->z; //

					  //Transform: rotate
	*eye = TransformVector(*eye, Rotate(degrees, upaxis));
	*center = TransformVector(*center, Rotate(degrees, upaxis));
	*up = TransformVector(*up, Rotate(degrees, upaxis));
	*side = TransformVector(*side, Rotate(degrees, upaxis));

	//Translate back to the original position
	center->x += oldeye.x;
	center->y += oldeye.y;
	center->z += oldeye.z;
	up->x += oldeye.x;
	up->y += oldeye.y;
	up->z += oldeye.z;
	side->x += oldeye.x;
	side->y += oldeye.y;
	side->z += oldeye.z;
	eye->x += oldeye.x; //
	eye->y += oldeye.y; // :)
	eye->z += oldeye.z; //
}

void TurnUp(float degrees, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	//In Crystal Ball interface:
	//vec3 axis = normalize(cross(eye, up));
	//eye = eye * rotate(degrees, axis);
	//up = up * rotate(degrees, axis);
	/*
	Vector3D axis = Normalize(CrossProduct(*eye, *up));
	*eye = TransformVector(*eye, Rotate(degrees, axis));
	*up = TransformVector( *up, Rotate(degrees, axis));
	*/

	//Set up axis
	Vector3D sideaxis;
	sideaxis.x = side->x - eye->x;
	sideaxis.y = side->y - eye->y;
	sideaxis.z = side->z - eye->z;
	sideaxis = Normalize(sideaxis);

	Vector3D oldeye;
	oldeye.x = eye->x;
	oldeye.y = eye->y;
	oldeye.z = eye->z;

	//Translate to world origin
	center->x -= eye->x;
	center->y -= eye->y;
	center->z -= eye->z;
	up->x -= eye->x;
	up->y -= eye->y;
	up->z -= eye->z;
	side->x -= eye->x;
	side->y -= eye->y;
	side->z -= eye->z;
	eye->x -= eye->x; //
	eye->y -= eye->y; // :)
	eye->z -= eye->z; //

					  //Transform (rotate)
	*eye = TransformVector(*eye, Rotate(degrees, sideaxis));
	*center = TransformVector(*center, Rotate(degrees, sideaxis));
	*up = TransformVector(*up, Rotate(degrees, sideaxis));
	*side = TransformVector(*side, Rotate(degrees, sideaxis));

	//Translate back
	center->x += oldeye.x;
	center->y += oldeye.y;
	center->z += oldeye.z;
	up->x += oldeye.x;
	up->y += oldeye.y;
	up->z += oldeye.z;
	side->x += oldeye.x;
	side->y += oldeye.y;
	side->z += oldeye.z;
	eye->x += oldeye.x;
	eye->y += oldeye.y;
	eye->z += oldeye.z;
}

void RollLeft(float degrees, Vector3D *eye, Vector3D *center, Vector3D *up, Vector3D *side)
{
	//In Crystal Ball interface:
	//vec3 axis = normalize(center - eye);
	//up = up * rotate(degrees, axis);
	/*
	Vector3D axis;
	axis.x = center.x - eye.x;
	axis.y = center.y - eye.y;
	axis.z = center.z - eye.z;
	axis = Normalize(axis);
	*up = TransformVector(*up, Rotate(degrees, axis));
	*/

	//Set up rotation axis
	Vector3D frontaxis;
	frontaxis.x = center->x - eye->x;
	frontaxis.y = center->y - eye->y;
	frontaxis.z = center->z - eye->z;
	frontaxis = Normalize(frontaxis);

	Vector3D oldeye;
	oldeye.x = eye->x;
	oldeye.y = eye->y;
	oldeye.z = eye->z;

	//Translate to world origin
	center->x -= eye->x;
	center->y -= eye->y;
	center->z -= eye->z;
	up->x -= eye->x;
	up->y -= eye->y;
	up->z -= eye->z;
	side->x -= eye->x;
	side->y -= eye->y;
	side->z -= eye->z;
	eye->x -= eye->x;
	eye->y -= eye->y;
	eye->z -= eye->z;

	//Transform (rotate)
	*eye = TransformVector(*eye, Rotate(degrees, frontaxis));
	*center = TransformVector(*center, Rotate(degrees, frontaxis));
	*up = TransformVector(*up, Rotate(degrees, frontaxis));
	*side = TransformVector(*side, Rotate(degrees, frontaxis));

	//Translate back
	center->x += oldeye.x;
	center->y += oldeye.y;
	center->z += oldeye.z;
	up->x += oldeye.x;
	up->y += oldeye.y;
	up->z += oldeye.z;
	side->x += oldeye.x;
	side->y += oldeye.y;
	side->z += oldeye.z;
	eye->x += oldeye.x;
	eye->y += oldeye.y;
	eye->z += oldeye.z;
}

void Move(float move_amount, Camera* camera, MovementCallback Movement);

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, INT iCmdShow)
{
	HWND		hwnd;
	HACCEL		hAccel;
	MSG		msg;
	WNDCLASS	wndclass;
	char		szAppName[] = TEXT("Voxel");

	wndclass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, "engine");
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = NULL;
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Unable to register window class."), szAppName, MB_ICONERROR);
		return 0;
	}
	hwnd = CreateWindow(szAppName,			// window class name
		TEXT("Voxel"),		// window caption
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS,		// window style
		CW_USEDEFAULT,			// initial x position
		CW_USEDEFAULT,			// initial y position
		CW_USEDEFAULT,			// initial x size
		CW_USEDEFAULT,			// initial y size
		NULL,				// parent window handle
		NULL,				// window menu handle
		hInstance,			// program instance handle
		NULL);				// creation parameters

	hAccel = LoadAccelerators(hInstance, szAppName);
	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (TRUE)
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if (GetMessage(&msg, NULL, 0, 0) > 0)
			{
				if (!TranslateAccelerator(hwnd, hAccel, &msg))
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
			else {
				break;
			}
		}
		else
		{
			//			EngineProc(E_RENDER, 0, 0);
		}
	}

	if (!UnregisterClass(szAppName, hInstance))
	{
		MessageBox(NULL, TEXT("Unable to unregister class."), szAppName, MB_ICONERROR);
		return 1;
	}

	return (int)msg.wParam;
}

void draw_pixels(HDC hdc, HDC hdcMem, int width, int height)
{
	HBITMAP hBitmap, hOldBitmap;

	hBitmap = CreateCompatibleBitmap(hdc, width, height);
	SetBitmapBits(hBitmap, sizeof(int) * width * height, vram);
	hdcMem = CreateCompatibleDC(hdc);
	hOldBitmap = (HBITMAP)SelectObject(hdcMem, hBitmap);

	// This scaling is a little strange because Stretch maintains aspect ratios
	StretchBlt(hdc, 0, 0, cxClient, cyClient, hdcMem, 0, 0, width, height, SRCCOPY);
	SelectObject(hdcMem, hOldBitmap);
	DeleteDC(hdcMem);
	DeleteObject(hBitmap);
}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static HDC		hdc;
	static HGLRC		hglrc;
	static RECT rect;

	static HDC	hdcMem;
	static int	old_style = 0;
	static int	new_style = WS_CHILD | WS_VISIBLE;
	static int	xres, yres;
	PAINTSTRUCT	ps;



	switch (message)
	{
	case WM_CREATE:
	{
		HMONITOR hmon;
		MONITORINFO mi = { sizeof(MONITORINFO) };

		xres = screen_width;
		yres = screen_height;
		SetTimer(hwnd, WM_TICK, 16, NULL);

		hmon = MonitorFromWindow(hwnd, MONITOR_DEFAULTTONEAREST);
		GetMonitorInfo(hmon, &mi);

		SetTimer(hwnd, WM_TICK, 16, NULL);

		vram = new unsigned char[4 * screen_width * screen_height];
		memset(vram, 0, 4 * screen_width * screen_height);


		if (false == LoadVxl("./data/seed7t1.vxl"))
		{
			MessageBox(hwnd, "Error loading VXL.\n", "Error", 0);
			return 0;
		}

		return 0;
	}
	case WM_MOUSEMOVE:
	{
		int	x, y;

		x = LOWORD(lParam);
		y = HIWORD(lParam);
	}
	return 0;

	case WM_TIMER:
	{
		int width = cxClient;
		int height = cyClient;

		InvalidateRect(hwnd, &rect, FALSE);
		return 0;
	}

	case WM_KEYDOWN:

		switch (wParam)
		{
		case VK_ESCAPE:
				exit(0);
				break;
		case VK_LEFT:
			{
				Move(degleft, player, TurnLeft);
			}
			break;
		case VK_RIGHT:
			{
				Move(-degleft, player, TurnLeft);
			}
			break;
		case VK_UP:
			{
				Move(degup, player, TurnUp);
			}
			break;
		case VK_DOWN:
			{
				Move(-degup, player, TurnUp);
			}
			break;
		case VK_SPACE:
			{
				Move(stepy, player, MoveY);
			}
			break;
		case VK_CONTROL:
			{
				Move(-stepy, player, MoveY);
			}
			break;
		case 'W':
			{
				Move(stepz, player, MoveZ);
			}
			break;
		case 'S':
			{
				Move(-stepz, player, MoveZ);
			}
			break;
		case 'A':
			{
				Move(-stepx, player, MoveX);
			}
			break;
		case 'D':
			{
				Move(stepx, player, MoveX);
			}
			break;
		case 'O':
			{
				Move(-degroll, player, RollLeft);
			}
			break;
		case 'P':
			{
				Move(degroll, player, RollLeft);
			}
			break;

		case 'X':
			{
				VoxelLaserKill();
			}
			break;
		}

		return 0;

	case WM_SIZE:
	{
		int	width, height;

		width = LOWORD(lParam);
		height = HIWORD(lParam);

		cxClient = width;
		cyClient = height;
		rect.right = cxClient;
		rect.bottom = cyClient;

		InvalidateRect(hwnd, NULL, false);
	}
	return 0;
	case WM_PAINT:
		hdc = BeginPaint(hwnd, &ps);
		DrawVxlRaycast(vram, screen_width, screen_height);
		draw_pixels(hdc, hdcMem, xres, yres);
		EndPaint(hwnd, &ps);
		return 0;

	case WM_SYSCOMMAND:
		switch (wParam)
		{
		case SC_SCREENSAVE:
		case SC_MONITORPOWER:
			return 0;
		}
		break;

	case WM_DESTROY:
		if (NULL != vram)
		{
			delete[] vram;
			vram = NULL;
		}

		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}




inline float DotProduct(const Vector3D &v1, const Vector3D &v2)
{
	return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
}

Vector3D CrossProduct(const Vector3D &v1, const Vector3D &v2)
{
	Vector3D result;
	result.x = v1.y*v2.z - v1.z*v2.y;
	result.y = v1.z*v2.x - v1.x*v2.z;
	result.z = v1.x*v2.y - v1.y*v2.x;
	return result;
}

// interpolates movement between old and new position calling callback function each time (smooths movement a bit
void Move(float move_amount, Camera* camera, MovementCallback Movement)
{
	const int number_of_move_fractions = 10;
	float move_amount_inc = move_amount / number_of_move_fractions;
	float abs_move_amount = fabsf(move_amount);
	float abs_move_amount_inc = fabsf(move_amount_inc);
	for (float move_amount_tmp = 0.0f; move_amount_tmp <= abs_move_amount; move_amount_tmp += abs_move_amount_inc)
	{
		Movement(move_amount_inc, &(camera->eye), &(camera->center), &(camera->up), &(camera->side));
	}
}

//(Ken Silverman):
//Called at least once for every voxel of the board
//issolid: 0 - air, 1 - solid
void setgeom(long x, long y, long z, long issolid)
{
	long index = (256 * (1024 * x + y) + z);
	if (issolid)
	{
		bitarray[index / 8] |= (1 << (index % 8));
	}
	else
	{
		bitarray[index / 8] &= ~(1 << (index % 8));
	}
}

long getgeom(long x, long y, long z)
{
	long index = (256 * (1024 * x + y) + z);
	return bitarray[index / 8] & (1 << (index % 8));
}

//(Ken Silverman):
//Called only for surface voxels
//A surface voxel is any solid voxel with at least 1 air voxel
//  on one of its 6 sides. All solid voxels at z = 0 are automatically
//  surface voxels, but this is not true for x = 0, x = 1023, y = 0, y = 1023,
//  z = 255 (I believe)
//argb: 32-bit color, high byte is used for shading scale (can be ignored)
bool setcol(long x, long y, long z, long argb)
{
	Voxel* voxel = new Voxel();
	if (NULL == voxel)
	{
		return false;
	}
	voxel->x = x;
	voxel->y = y;
	voxel->z = z;
	voxel->argb = argb;
	long index = 256 * (1024 * x + y) + z;
	colorarray[index] = voxel;
	return true;
}

// Reads voxel file
bool LoadVxl(char* filename)
{
	FILE* file;
	long i = 0;
	unsigned char *v = NULL, *vbuf = NULL;
	unsigned int file_size = 0;

	memset(bitarray, 0, kBitArraySize);

	file = fopen(filename, "rb");
	if (NULL == file)
	{
		//ShowMessage("LoadVxl() file open fail!\n");
		return false;
	}

	fseek(file, 0, SEEK_END);
	file_size = ftell(file);
	fseek(file, 0, SEEK_SET);


	fread(&i, 4, 1, file);
	if (i != 0x09072000)
	{
		//ShowMessage("LoadVxl() signature reading fail!\n");
		fclose(file);
		return false;
	}

	fread(&i, 4, 1, file);
	if (1024 != i)
	{
		//ShowMessage("LoadVxl() X-dimension read fail! Expected 1024, read ");
		fclose(file);
		return false;
	}

	fread(&i, 4, 1, file);
	if (1024 != i)
	{
		//ShowMessage("LoadVxl() Y-dimension read fail! Expected 1024, read ");
		fclose(file);
		return false;
	}

	fread(&ipos, 24, 1, file); //camera position
	fread(&istr, 24, 1, file); //unit right vector
	fread(&ihei, 24, 1, file); //unit down vector
	fread(&ifor, 24, 1, file); //unit forward vector



	i = file_size - ftell(file);
	vbuf = new unsigned char[i];
	if (NULL == vbuf)
	{
		//ShowMessage("LoadVxl() V-buffer allocation fail!\n");
		fclose(file);
		return false;
	}

	fread(vbuf, i, 1, file);
	fclose(file);

	//  1024 * 1024 * 256 * 16 bytes = 4GB
	// actually array of pointers, should eliminate
	colorarray = (Voxel **)calloc(1024 * 1024 * 256 * sizeof(Voxel *), 1);

	memset(bitarray, 0xFF, 33554432);

	v = vbuf;
	for (long y = 0; y < 1024; ++y)
	{
		for (long x = 0; x < 1024; ++x)
		{
			long z = 0;
			while (1)
			{
				for (long i = z; i < v[1]; ++i)
				{
					setgeom(x, y, i, 0);
				}

				for (long z = v[1]; z <= v[2]; ++z)
				{
					bool result = setcol(x, y, z, *(long *)&v[(z - v[1] + 1) << 2]);
					if (false == result)
						return result;
				}

				if (!v[0])
				{
					break;
				}

				z = v[2] - v[1] - v[0] + 2;
				v += v[0] << 2;

				for (z += v[3]; z<v[3]; ++z)
				{
					bool result = setcol(x, y, z, *(long *)&v[(z - v[3]) << 2]);
					if (false == result)
						return result;
				}
			}
			v += ((((long)v[2]) - ((long)v[1]) + 2) << 2);
		}
	}
	delete[] vbuf;
	return true;
}

// sets pixel sx, sy to color with fade to black effect
void PixelSet(const long sx, const long sy, unsigned long argb, float fade, unsigned char *buffer, int width, int height)
{
	if ((sx < 0) || (sx > width - 1) || (sy < 0) || (sy > height - 1))
	{
		return;
	}

	long index = width * sy + sx;
	unsigned char red = (unsigned char)(argb & 0x000000FF);
	unsigned char grn = (unsigned char)((argb & 0x0000FF00) >> 8);
	unsigned char blu = (unsigned char)((argb & 0x00FF0000) >> 16);

	buffer[4 * index + 0] = (unsigned char)(red * fade);
	buffer[4 * index + 1] = (unsigned char)(grn * fade);
	buffer[4 * index + 2] = (unsigned char)(blu * fade);
	buffer[4 * index + 3] = 255;
}

// Fills pixel sx,sy with voxel color of indexed voxel
void DrawVoxelSample(long sx, long sy, float fade, Voxel* voxel, long index)
{
	if (NULL == voxel)
		return;
	PixelSet(sx, sy, voxel->argb, fade, vram, screen_width, screen_height);
}

// Removes a sphere from where hit was detected
void ShootVoxel(long sx, long sy, float fade, Voxel* voxel, long index)
{
	if (NULL == voxel)
		return;
	unsigned long voxel_paintcolor = voxel->argb;
	setgeom(voxel->x, voxel->y, voxel->z, 0);

	const long x0 = voxel->x;
	const long y0 = voxel->y;
	const long z0 = voxel->z;
	const long radius = 10;
	for (long x = x0 - 2 * radius; x < x0 + 2 * radius; ++x)
	{
		for (long y = y0 - 2 * radius; y < y0 + 2 * radius; ++y)
		{
			for (long z = voxel->z - 2 * radius; z < voxel->z + 2 * radius; ++z)
			{
				if ((x < 0) || (x >= 1024) || (y < 0) || (y >= 1024) ||
					(z < 0) && (z >= 256))
				{
					continue;
				}

				long current_index = 256 * (1024 * x + y) + z;

				if (current_index == index)
					continue;

				long sphere = (x - x0) * (x - x0) +
					(y - y0) * (y - y0) +
					(z - z0) * (z - z0);

				if (sphere < radius * radius)
				{
					setgeom(x, y, z, 0);
					if (NULL != colorarray[current_index])
					{
						delete colorarray[current_index];
						colorarray[current_index] = NULL;
					}
				}
				else if ((sphere >= radius * radius) &&
					(sphere <= (radius + radius / 4) *
					(radius + radius / 4)))
				{
					// TODO: Fill gaps
					//setgeom(x, y, z, 1);
					//setcol(x, y, z, voxel_paintcolor);

					// Shell variant
					if (0 != getgeom(x, y, z))
					{
						Voxel* current_voxel = colorarray[current_index];
						if (NULL != current_voxel)
						{
							current_voxel->argb = voxel_paintcolor;
						}
					}
				}
			}
		}
	}

	colorarray[index] = NULL;
	if (NULL != voxel)
	{
		delete voxel;
		voxel = NULL;
	}
}

// Computes voxel index and calls callback on the voxel
// sx = screenx, sy = screeny
bool VisitVoxel(long sx, long sy, long x, long y, long z, float fade, ProcessVoxelCallback ProcessVoxel)
{
	if ((x < 0) || (x >= 1024) || (y < 0) || (y >= 1024) || (z < 0) && (z >= 256))
	{
		return false;
	}

	if (0 == getgeom(x, y, z))
		return false;

	long index = 256 * (1024 * x + y) + z;

	if (index < 0 || index >= 1024 * 1024 * 256)
		return false;

	if (NULL == colorarray[index])
		return false;
	Voxel* voxel = colorarray[index];
	ProcessVoxel(sx, sy, fade, voxel, index);
	return true;
}

// returns sign of X (is division really faster than a branch?)
long sgn(long x)
{
	if (0 == x)
		return 1;
	return x / abs(x);
}

// 3d DDA Line function
void Line(long scrx, long scry, long x, long y, long z,
	long dx, long dy, long dz,
	VisitVoxelCallback VisitVoxel,
	ProcessVoxelCallback ProcessVoxel)
{
	long n, sx, sy, sz, exy, exz, ezy, ax, ay, az, bx, by, bz;

	sx = sgn(dx); sy = sgn(dy); sz = sgn(dz);
	ax = abs(dx); ay = abs(dy); az = abs(dz);
	bx = 2 * ax;  by = 2 * ay;  bz = 2 * az;
	exy = ay - ax;
	exz = az - ax;
	ezy = ay - az;

	bool hit = false;
	long dist = ax + ay + az;
	n = dist;
	while (--n)
	{
		float fade = (float)n / (float) (dist);

		bool hit = VisitVoxel(scrx, scry, x, y, z, fade, ProcessVoxel);
		if (true == hit)
			break;
		if (exy < 0)
		{
			if (exz < 0)
			{
				x += sx;
				exy += by;
				exz += bz;
			}
			else
			{
				z += sz;
				exz -= bx;
				ezy += by;
			}
		}
		else
		{
			if (ezy < 0)
			{
				z += sz;
				exz -= bx;
				ezy += by;
			}
			else
			{
				y += sy;
				exy -= bx;
				ezy -= bz;
			}
		}
	}
}

void SetupLCS(const Camera& camera, Vector3D* up, Vector3D* front, Vector3D* right)
{
	up->x = camera.up.x - camera.eye.x;
	up->y = camera.up.y - camera.eye.y;
	up->z = camera.up.z - camera.eye.z;

	front->x = camera.center.x - camera.eye.x;
	front->y = camera.center.y - camera.eye.y;
	front->z = camera.center.z - camera.eye.z;

	right->x = camera.side.x - camera.eye.x;
	right->y = camera.side.y - camera.eye.y;
	right->z = camera.side.z - camera.eye.z;
}

void SetupViewport(const Camera& camera,
	const Vector3D& up, const Vector3D& front, const Vector3D& right,
	const float d1, const float d2,
	Vector3D viewport[3],
	Vector3D* viewport_dxvector, Vector3D* viewport_dyvector)
{
	viewport[0].x = camera.eye.x + d1 * (up.x - right.x);
	viewport[0].y = camera.eye.y + d1 * (up.y - right.y);
	viewport[0].z = camera.eye.z + d1 * (up.z - right.z);

	viewport[1].x = camera.eye.x + d1 * (up.x + right.x);
	viewport[1].y = camera.eye.y + d1 * (up.y + right.y);
	viewport[1].z = camera.eye.z + d1 * (up.z + right.z);

	viewport[2].x = camera.eye.x - d1 * (up.x + right.x);
	viewport[2].y = camera.eye.y - d1 * (up.y + right.y);
	viewport[2].z = camera.eye.z - d1 * (up.z + right.z);

	for (int i = 0; i < 3; ++i)
	{
		viewport[i].x += d2 * front.x;
		viewport[i].y += d2 * front.y;
		viewport[i].z += d2 * front.z;
		viewport[i].w = 1.0f;
	}

	viewport_dxvector->x = viewport[1].x - viewport[0].x;
	viewport_dxvector->y = viewport[1].y - viewport[0].y;
	viewport_dxvector->z = viewport[1].z - viewport[0].z;
	viewport_dxvector->w = 1.0f;

	viewport_dyvector->x = viewport[2].x - viewport[0].x;
	viewport_dyvector->y = viewport[2].y - viewport[0].y;
	viewport_dyvector->z = viewport[2].z - viewport[0].z;
	viewport_dyvector->w = 1.0f;
}


// Calls Line() passing callback ShootVoxel()
void VoxelLaserKill()
{
	SetupLCS(*player, &up, &front, &right);

	Vector3D raystart, rayend;
	raystart.x = player->eye.x;
	raystart.y = player->eye.y;
	raystart.z = player->eye.z;
	raystart.w = 1.0f;
	rayend.x = raystart.x + front.x;
	rayend.y = raystart.y + front.y;
	rayend.z = raystart.z + front.z;
	rayend.w = 1.0f;

	Voxel voxel_raystart, voxel_rayend, voxel_ray;
	voxel_raystart.x = (long)((raystart.x - boardorigin.x) / h);
	voxel_raystart.y = (long)((raystart.y - boardorigin.y) / h);
	voxel_raystart.z = (long)((raystart.z - boardorigin.z) / h);
	voxel_rayend.x = (long)((rayend.x - boardorigin.x) / h);
	voxel_rayend.y = (long)((rayend.y - boardorigin.y) / h);
	voxel_rayend.z = (long)((rayend.z - boardorigin.z) / h);
	voxel_ray.x = voxel_rayend.x - voxel_raystart.x;
	voxel_ray.y = voxel_rayend.y - voxel_raystart.y;
	voxel_ray.z = voxel_rayend.z - voxel_raystart.z;
	voxel_ray.x *= kRange;
	voxel_ray.y *= kRange;
	voxel_ray.z *= kRange;

	Line(screen_width / 2, screen_height / 2,
		voxel_raystart.x, voxel_raystart.y, voxel_raystart.z,
		voxel_ray.x, voxel_ray.y, voxel_ray.z,
		VisitVoxel, ShootVoxel);
}


// Main drawing function, calls Line() passing callback DrawVoxelSample()
void DrawVxlRaycast(unsigned char *buffer, int width, int height)
{
	if (NULL == vram)
		return;

	SetupLCS(*player, &up, &front, &right);
	SetupViewport(*player, up, front, right, d1, d2, viewport, &viewport_dxvector, &viewport_dyvector);

	Vector3D raystart;
	raystart.x = player->eye.x;
	raystart.y = player->eye.y;
	raystart.z = player->eye.z;
	raystart.w = 1.0f;

	Voxel voxel_raystart;
	voxel_raystart.x = (long)((raystart.x - boardorigin.x) / h);
	voxel_raystart.y = (long)((raystart.y - boardorigin.y) / h);
	voxel_raystart.z = (long)((raystart.z - boardorigin.z) / h);


	memset(buffer, 0, 4 * width * height);
	for (long y = 0; y < height; ++y)
	{
		#pragma omp parallel
		#pragma omp for
		for (long x = 0; x < width; ++x)
		{
			Vector3D rayend;
			rayend.x = viewport[0].x + x * ddx * viewport_dxvector.x + +y * ddy * viewport_dyvector.x;
			rayend.y = viewport[0].y + x * ddx * viewport_dxvector.y + +y * ddy * viewport_dyvector.y;
			rayend.z = viewport[0].z + x * ddx * viewport_dxvector.z + +y * ddy * viewport_dyvector.z;
			rayend.w = 1.0f;

			Voxel voxel_rayend;
			voxel_rayend.x = (long)((rayend.x - boardorigin.x) / h);
			voxel_rayend.y = (long)((rayend.y - boardorigin.y) / h);
			voxel_rayend.z = (long)((rayend.z - boardorigin.z) / h);

			Voxel voxel_ray;
			voxel_ray.x = voxel_rayend.x - voxel_raystart.x;
			voxel_ray.y = voxel_rayend.y - voxel_raystart.y;
			voxel_ray.z = voxel_rayend.z - voxel_raystart.z;
			voxel_ray.x *= kFar;
			voxel_ray.y *= kFar;
			voxel_ray.z *= kFar;
			Line(x, y, voxel_raystart.x, voxel_raystart.y, voxel_raystart.z,
				voxel_ray.x, voxel_ray.y, voxel_ray.z,
				VisitVoxel, DrawVoxelSample);
		}
	}
}
